import os
import tempfile
import configparser
import logging

ROOKI_HOME = os.path.abspath(os.path.dirname(__file__))
DEFAULT_CFG = os.path.join(ROOKI_HOME, "default.cfg")

RAW_OPTIONS = [
    ("logging", "format"),
]

CONFIG = None
LOGGER = logging.getLogger("ROOKI")


def get_config_value(section, option):
    """Get desired value from  configuration files
    :param section: section in configuration files
    :type section: string
    :param option: option in the section
    :type option: string
    :returns: value found in the configuration file
    """

    if not CONFIG:
        load_configuration()

    value = ""

    if CONFIG.has_section(section):
        if CONFIG.has_option(section, option):
            raw = (section, option) in RAW_OPTIONS
            value = CONFIG.get(section, option, raw=raw)

            # Convert Boolean string to real Boolean values
            if value.lower() == "false":
                value = False
            elif value.lower() == "true":
                value = True

    return value


def load_configuration(cfgfiles=None):
    """Load rooki configuration from configuration files.
    The later configuration file in the array overwrites configuration
    from the first.
    :param cfgfiles: list of configuration files
    """
    global CONFIG

    LOGGER.info("loading configuration")
    CONFIG = configparser.ConfigParser(os.environ)

    LOGGER.debug("setting default values")
    CONFIG.add_section("service")
    CONFIG.set("service", "url", "http://localhost:5000/wps")
    CONFIG.set("service", "output_dir", tempfile.gettempdir())

    config_files = _get_default_config_files_location()
    if cfgfiles:
        config_files.extend(cfgfiles)

    if "ROOKI_CFG" in os.environ:
        config_files.append(os.environ["ROOKI_CFG"])

    loaded_files = CONFIG.read(config_files)
    if loaded_files:
        LOGGER.info("Configuration file(s) {} loaded".format(loaded_files))
    else:
        LOGGER.info("No configuration files loaded. Using default values")


def _get_default_config_files_location():
    """Get the locations of the standard configuration files. These are
        1. `$ROOKI_HOME/default.cfg`
        2. `/etc/rooki.cfg`
        3. `$HOME/.rooki.cfg`
    :returns: configuration files
    :rtype: list of strings
    """
    LOGGER.debug("trying to estimate the default location")
    cfgfiles = [
        DEFAULT_CFG,
        "/etc/rooki.cfg",
    ]
    if "HOME" in os.environ:
        cfgfiles.append(os.path.join(os.environ["HOME"], ".rooki.cfg"))
    return cfgfiles



import os
import tempfile

from birdy import WPSClient
from owslib.wps import ASYNC
from rooki import config
from rooki.results import Result


import logging


class Rooki(WPSClient):
    def __init__(
        self,
        url=None,
        mode=None,
        verify=None,
        output_dir=None,
        headers=None,
        lineage=True,
    ):
        self._url = url or os.environ.get(
            "ROOK_URL", config.get_config_value("service", "url")
        )
        self._mode = mode or os.environ.get(
            "ROOK_MODE", config.get_config_value("service", "mode")
        )
        if verify is None:
            self._verify = os.environ.get(
                "ROOK_SSL_VERIFY", config.get_config_value("service", "ssl_verify")
            )
        else:
            self._verify = verify
        self._output_dir = output_dir or os.environ.get(
            "ROOKI_OUTPUT_DIR", config.get_config_value("service", "output_dir")
        )
        progress = self.mode == ASYNC
        if "ACCESS_TOKEN" in os.environ:
            access_token = os.environ["ACCESS_TOKEN"]
            if headers is None:
                headers = {}
            headers["Authorization"] = f"Bearer {access_token}"
        super(Rooki, self).__init__(
            self._url,
            verify=self._verify,
            progress=progress,
            headers=headers,
            lineage=lineage,
        )
        self._notebook = False
        self.logger = logging.getLogger("rooki")
        self.logger.addHandler(logging.NullHandler())

    def cancel(self):
        # https://github.com/roocs/rook/issues/143
        self.logger.warn("cancel was called but is not available.")

    @property
    def url(self):
        return self._url

    @property
    def mode(self):
        return self._mode

    @property
    def verify(self):
        return self._verify

    @property
    def output_dir(self):
        if not self._output_dir:
            self._output_dir = tempfile.gettempdir()
        elif not os.path.isdir(self._output_dir):
            os.makedirs(self._output_dir)
        return self._output_dir

    def _execute(self, pid, **kwargs):
        try:
            resp = super(Rooki, self)._execute(pid, **kwargs)
        except Exception:
            raise RuntimeError("execution failed")
        return Result(resp, output_dir=self.output_dir)



"""A mock-up rooki API."""

import json
from collections import defaultdict

from .client import Rooki

rooki = Rooki()


def reinit():
    global rooki
    rooki = Rooki()


class Operator:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def _tree(self, tree=defaultdict(dict)):
        # args = [arg._tree(tree) for arg in self.args]
        [arg._tree(tree) for arg in self.args]
        tree["steps"][self.method_key] = {
            "run": self.METHOD,
            "in": {
                "collection": _unpack_if_single([arg.collection for arg in self.args]),
                **self.kwargs,
            },
        }
        tree["outputs"]["output"] = self.collection
        return tree

    def _serialise(self, doc="workflow"):
        tree = self._tree()
        tree["doc"] = doc
        return json.dumps(tree)

    def orchestrate(self):
        reinit()
        return rooki.orchestrate(workflow=self._serialise())

    @property
    def method_key(self):
        methods = self._get_methods([])
        return f"{self.METHOD}_{self.variable}_{methods.count(self.METHOD)}"

    def _get_methods(self, methods):
        # method_names = [arg._get_methods(methods) for arg in self.args]
        [arg._get_methods(methods) for arg in self.args]
        methods.append(self.METHOD)
        return methods

    @property
    def collection(self):
        return f"{self.method_key}/output"

    @property
    def variable(self):
        return _unpack_if_single([arg.variable for arg in self.args])


class Input:
    def __init__(self, variable, dataset):
        self.variable = variable
        self.dataset = dataset

    def _serialise(self, doc="workflow"):
        tree = self._tree()
        tree["doc"] = doc
        return json.dumps(tree)

    def orchestrate(self):
        reinit()
        return rooki.orchestrate(workflow=self._serialise())

    def _tree(self, tree=defaultdict(dict)):
        tree["inputs"][self.variable] = self.dataset
        return tree

    def _get_methods(self, methods):
        return methods

    @property
    def collection(self):
        return f"inputs/{self.variable}"


class Average(Operator):
    METHOD = "average"


class AverageByTime(Operator):
    METHOD = "average_time"


class Subset(Operator):
    METHOD = "subset"


class Diff(Operator):
    """This is a dummy operator."""

    METHOD = "diff"

    def _tree(self, tree=defaultdict(dict)):
        # args = [arg._tree(tree) for arg in self.args]
        [arg._tree(tree) for arg in self.args]
        tree["steps"][self.method_key] = {
            "run": self.METHOD,
            "in": {
                "collection_a": _unpack_if_single([self.args[0].collection]),
                "collection_b": _unpack_if_single([self.args[1].collection]),
                **self.kwargs,
            },
        }
        tree["outputs"]["output"] = self.collection
        return tree

    @property
    def variable(self):
        return "var"


def _unpack_if_single(list):
    try:
        [item] = list
    except ValueError:
        item = list
    return item



# -*- coding: utf-8 -*-

"""Top-level package for rooki."""

from rooki.client import Rooki

# from rooki import operators


rooki = Rooki()


def reinit():
    global rooki
    rooki = Rooki()


__all__ = [
    "rooki",
    "Rooki",
    "reinit",
    # "operators",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""

from setuptools import setup, find_packages

__author__ = "Carsten Ehbrecht"
__contact__ = "ehbrecht@dkrz.de"
__copyright__ = "Copyright 2018 United Kingdom Research and Innovation"
__license__ = "BSD - see LICENSE file in top-level package directory"
__version__ = "0.4.0"

# One strategy for storing the overall version is to put it in the top-level
# package's __init__ but Nb. __init__.py files are not needed to declare
# packages in Python 3
# from rooki import __version__ as _package_version

# Populate long description setting with content of README
#
# Use markdown format read me file as GitHub will render it automatically
# on package page
with open("README.rst") as readme_file:
    _long_description = readme_file.read()


requirements = [line.strip() for line in open("requirements.txt")]
test_requirements = [
    "pytest",
]

setup_requirements = [
    "pytest-runner",
]


setup(
    author=__author__,
    author_email=__contact__,
    # See:
    # https://www.python.org/dev/peps/pep-0301/#distutils-trove-classification
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Environment :: Web Environment",
        "Intended Audience :: End Users/Desktop",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Topic :: Security",
        "Topic :: Internet",
        "Topic :: Scientific/Engineering",
        "Topic :: System :: Distributed Computing",
        "Topic :: System :: Systems Administration :: Authentication/Directory",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    description="A client for roocs climate data operations service.",
    license=__license__,
    # This qualifier can be used to selectively exclude Python versions -
    # in this case early Python 2 and 3 releases
    python_requires=">=3.6.0",
    entry_points={
        "console_scripts": [
            "rooki=rooki.cli:main",
        ],
    },
    install_requires=[
        requirements,
        # "pymetalink @ git+https://github.com/metalink-dev/pymetalink.git",
    ],
    long_description=_long_description,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="rooki",
    name="rooki",
    packages=find_packages(include=["rooki"]),
    setup_requires=setup_requirements,
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/roocs/rooki",
    version=__version__,
    zip_safe=False,
)



import tempfile
import requests
from bs4 import BeautifulSoup

import logging


class Result(object):
    def __init__(self, response, output_dir=None, verify=False):
        self.response = response
        self.output_dir = tempfile.mkdtemp(prefix="metalink_", dir=output_dir)
        self.verify = verify
        # cache
        self._xml = None
        self._doc = None
        self._size = None
        self._num_files = None

    @property
    def ok(self):
        return self.response.isSucceded()

    @property
    def status(self):
        if self.ok:
            status = self.response.status
        else:
            status = "; ".join([error.text.strip() for error in self.response.errors])
        return status

    @property
    def url(self):
        return self.response.get()[0]

    @property
    def xml(self):
        if not self._xml:
            try:
                self._xml = requests.get(self.url, verify=self.verify).text
            except Exception as e:
                raise Exception(f"Could not download metalink document. {e}")
        return self._xml

    @property
    def doc(self):
        if not self._doc:
            self._doc = BeautifulSoup(self.xml, "xml")
        return self._doc

    @property
    def size(self):
        """total size of all files in bytes."""
        if self._size is None:
            total_size = 0
            for size in self.doc.find_all("size"):
                total_size += int(size.text)
            self._size = total_size
        return self._size

    @property
    def size_in_mb(self):
        size_kb = self.size / 1024.0
        size_mb = size_kb / 1024.0
        return size_mb

    @property
    def size_in_gb(self):
        size_gb = self.size_in_mb / 1024.0
        return size_gb

    @property
    def num_files(self):
        if self._num_files is None:
            self._num_files = len(self.doc.find_all("file"))
        return self._num_files

    def download_urls(self):
        return [url.text for url in self.doc.find_all("metaurl")]

    def download(self):
        try:
            import metalink.download

            if self.verify is False:
                # TODO: should be handled in metalink
                import ssl

                ssl._create_default_https_context = ssl._create_unverified_context
            files = metalink.download.get(self.url, self.output_dir, segmented=False)
        except Exception as e:
            logging.error(f"Could not download files: {e}")
            files = []
        return files

    def datasets(self):
        try:
            import xarray as xr

            datasets = [xr.open_dataset(file) for file in self.download()]
        except Exception:
            datasets = []
        return datasets

    def provenance(self):
        return self.response.get()[1]

    def provenance_image(self):
        return self.response.get()[2]

    def __str__(self):
        if self.ok:
            msg = f"Metalink URL: {self.url}, num files: {self.num_files}"
        else:
            msg = self.status
        return msg

    def __repr__(self):
        return self.__str__()
